/*
 *  This file is part of ixemul.library for the Amiga.
 *  Copyright (C) 1991, 1992  Markus M. Wild
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define KERNEL
#include "ixemul.h"

#if __GNUC__ != 2
#define alloca __builtin_alloca
#endif

static int
__chmod_func (struct StandardPacket *sp, struct MsgPort *handler,
              BPTR parent_lock,
	      BSTR name,
	      int mask, int *no_error)
{
  sp->sp_Pkt.dp_Type = ACTION_SET_PROTECT;
  sp->sp_Pkt.dp_Arg1 = 0;
  sp->sp_Pkt.dp_Arg2 = parent_lock;
  sp->sp_Pkt.dp_Arg3 = name;
  sp->sp_Pkt.dp_Arg4 = mask;
  
  PutPacket (handler, sp);
  __wait_packet (sp);
  
  *no_error = sp->sp_Pkt.dp_Res1 == -1;
  return 1;
}


int
chmod(char *name, int mode)
{
  long amiga_mode = FIBF_READ|FIBF_WRITE|FIBF_EXECUTE|FIBF_DELETE;
  /* RWED-permissions are "lo-active", if cleared they allow the operation */
  int result;

  if (mode & S_IXUSR) amiga_mode &= ~FIBF_EXECUTE;
  if (mode & S_IWUSR) amiga_mode &= ~(FIBF_WRITE|FIBF_DELETE);
  if (mode & S_IRUSR) amiga_mode &= ~FIBF_READ;
  /* let a set sticky-bit mean that delete-permission is not granted */
  if (mode & S_ISVTX) amiga_mode |= FIBF_DELETE;

  result = __plock (name, __chmod_func, amiga_mode);

  return result == -1 ? 0 : -1;
}


/* a signal proof SetProtection(), nothing more ;-) */
int
achmod (char *name, int mode)
{
  return __plock (name, __chmod_func, mode) == -1 ? 0 : -1;
}
